/*
    %CopyrightBegin%

    SPDX-License-Identifier: Apache-2.0 OR LGPL-2.1-or-later

    Copyright (C) 2001-2004 Richard Carlsson
    Copyright Ericsson AB 2009-2025. All Rights Reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Alternatively, you may use this file under the terms of the GNU Lesser
    General Public License (the "LGPL") as published by the Free Software
    Foundation; either version 2.1, or (at your option) any later version.
    If you wish to allow use of your version of this file only under the
    terms of the LGPL, you should delete the provisions above and replace
    them with the notice and other provisions required by the LGPL; see
    <http://www.gnu.org/licenses/>. If you do not delete the provisions
    above, a recipient may use your version of this file under the terms of
    either the Apache License or the LGPL.

    %CopyrightEnd%
*/
/* standard EDoc style sheet */
body {
	font-family: Verdana, Arial, Helvetica, sans-serif;
      	margin-left: .25in;
       	margin-right: .2in;
       	margin-top: 0.2in;
       	margin-bottom: 0.2in;
       	color: #000000;
       	background-color: #ffffff;
}
h1,h2 {
 	margin-left: -0.2in;
}
div.navbar {
	background-color: #add8e6;
	padding: 0.2em;
}
h2.indextitle {
	padding: 0.4em;
	background-color: #add8e6;
}
h3.function,h3.typedecl {
	background-color: #add8e6;
 	padding-left: 1em;
}
div.spec {
 	margin-left: 2em;
	background-color: #eeeeee;
}
a.module {
	text-decoration:none
}
a.module:hover {
	background-color: #eeeeee;
}
ul.definitions {
	list-style-type: none;
}
ul.index {
	list-style-type: none;
	background-color: #eeeeee;
}

/*
 * Minor style tweaks
 */
ul {
	list-style-type: square;
}
table {
	border-collapse: collapse;
}
td {
	padding: 3
}
